"""Test output of database reports"""

from __future__ import absolute_import

import os
import unittest

from pycounter import csvhelper
from pycounter import report


class TestOutputDB1(unittest.TestCase):
    """Test output of DB1"""
    def setUp(self):
        filename = os.path.join(os.path.dirname(__file__), 'data/C4DB1.tsv')
        rep = report.parse(filename)
        with csvhelper.UnicodeReader(filename,
                                     delimiter='\t') as report_reader:
            self.file_content = list(report_reader)

        self.output_content = rep.as_generic()

    def test_header_content(self):
        self.assertEqual(self.file_content[0:7], self.output_content[0:7])

    def test_table_header(self):
        self.assertEqual(self.file_content[7], self.output_content[7])

    def test_totals(self):
        self.assertEqual(self.file_content, self.output_content)

    def test_data(self):
        for index, line in enumerate(self.file_content[9:], 9):
            self.assertEqual(line, self.output_content[index])


class TestOutputDB2(unittest.TestCase):
    """Test output of DB2"""
    def setUp(self):
        filename = os.path.join(os.path.dirname(__file__), 'data/C4DB2.tsv')
        rep = report.parse(filename)
        with csvhelper.UnicodeReader(filename,
                                     delimiter='\t') as report_reader:
            self.file_content = list(report_reader)

        self.output_content = rep.as_generic()

    def test_header_content(self):
        self.assertEqual(self.file_content[0:7], self.output_content[0:7])

    def test_table_header(self):
        self.assertEqual(self.file_content[7], self.output_content[7])

    def test_totals(self):
        self.assertEqual(self.file_content, self.output_content)

    def test_data(self):
        for index, line in enumerate(self.file_content[9:], 9):
            self.assertEqual(line, self.output_content[index])



"""Test COUNTER DB2 database report"""

from __future__ import absolute_import

import datetime
import os
import unittest

from pycounter.constants import METRICS
import pycounter.report


class ParseCounter4Example(unittest.TestCase):
    """Tests for parsing C4 DB2"""
    def setUp(self):
        self.report = pycounter.report.parse(
            os.path.join(os.path.dirname(__file__),
                         'data/C4DB2.tsv'))

    def test_reportname(self):
        self.assertEqual(self.report.report_type, u'DB2')
        self.assertEqual(self.report.report_version, 4)

    def test_year(self):
        self.assertEqual(self.report.year, 2012)

    def test_platform(self):
        for publication in self.report:
            self.assertEqual(publication.publisher, u"Megadodo Publications")
            self.assertEqual(publication.platform, u"HHGTTG Online")

    def test_stats(self):
        publication = self.report.pubs[0]
        self.assertEqual(
            [x[2] for x in publication],
            [0, 5, 0, 0, 0, 0])

    def test_report_metric(self):
        for metric in self.report.metric:
            self.assertTrue(metric in METRICS[self.report.report_type])

    def test_row_metric(self):
        publication = self.report.pubs[0]
        jan_data = next(iter(publication))
        self.assertEqual(jan_data[1], "Access denied: concurrent/"
                                      "simultaneous user license "
                                      "limit exceeded")

    def test_customer(self):
        self.assertEqual(self.report.customer,
                         u"University of Maximegalon")

    def test_date_run(self):
        self.assertEqual(self.report.date_run, datetime.date(2012, 7, 9))

    def test_period(self):
        self.assertEqual(self.report.period,
                         (datetime.date(2012, 1, 1),
                          datetime.date(2012, 6, 30)))



import datetime
import unittest


class TSVJR1(unittest.TestCase):
    report = None  # should be set in subclass setUp.

    def test_reportname(self):
        self.assertEqual(self.report.report_type, u'JR1')
        self.assertEqual(self.report.report_version, 4)

    def test_year(self):
        self.assertEqual(self.report.year, 2013)

    def test_platform(self):
        for publication in self.report:
            self.assertEqual(publication.publisher, u"aap")
            self.assertEqual(publication.platform,
                             u"Journal of Periodontology Online")

    def test_stats(self):
        publication = self.report.pubs[0]
        self.assertEqual([x[2] for x in publication],
                         [1, 4, 9, 4, 2, 0, 1, 4, 5, 2])
        publication = self.report.pubs[1]
        self.assertEqual([x[2] for x in publication],
                         [3, 0, 0, 0, 1, 0, 14, 1, 3, 5])

    def test_customer(self):
        self.assertEqual(self.report.customer, u"University of Maximegalon")

    def test_date_run(self):
        self.assertEqual(self.report.date_run, datetime.date(2013, 11, 22))

    def test_period(self):
        self.assertEqual(self.report.period,
                         (datetime.date(2013, 1, 1),
                          datetime.date(2013, 10, 31)))

    def test_dates(self):
        publication = self.report.pubs[0]
        self.assertEqual([x[0] for x in publication],
                         [datetime.date(2013, 1, 1),
                          datetime.date(2013, 2, 1),
                          datetime.date(2013, 3, 1),
                          datetime.date(2013, 4, 1),
                          datetime.date(2013, 5, 1),
                          datetime.date(2013, 6, 1),
                          datetime.date(2013, 7, 1),
                          datetime.date(2013, 8, 1),
                          datetime.date(2013, 9, 1),
                          datetime.date(2013, 10, 1),
                          ]
                         )



# coding: utf-8
"""Test parsing of COUNTER BR2 book report."""

from __future__ import absolute_import

import os
import unittest
import warnings

import pycounter.report


class ParseExample(unittest.TestCase):
    """Tests for parsing C3 BR2"""
    def setUp(self):
        self.report = pycounter.report.parse(
            os.path.join(os.path.dirname(__file__),
                         'data/simpleBR2.csv'))

    def test_reportname(self):
        self.assertEqual(self.report.report_type, u'BR2')
        self.assertEqual(self.report.report_version, 1)

    def test_year(self):
        self.assertEqual(self.report.year, 2012)

    def test_platform(self):
        for publication in self.report:
            self.assertEqual(publication.publisher, u"Megadodo Publications")
            self.assertEqual(publication.platform, u"HHGTTG Online")

    def test_stats(self):
        publication = self.report.pubs[0]
        self.assertEqual(
            [x[2] for x in publication],
            [0, 25, 0, 0, 0, 0])


class ParseCounter4Example(unittest.TestCase):
    """Tests for parsing C4 BR2"""
    def setUp(self):
        self.report = pycounter.report.parse(
            os.path.join(os.path.dirname(__file__),
                         'data/C4BR2.tsv'))

    def test_reportname(self):
        self.assertEqual(self.report.report_type, u'BR2')
        self.assertEqual(self.report.report_version, 4)

    def test_year(self):
        self.assertEqual(self.report.year, 2012)

    def test_platform(self):
        for publication in self.report:
            self.assertEqual(publication.publisher, u"Megadodo Publications")
            self.assertEqual(publication.platform, u"HHGTTG Online")

    def test_stats(self):
        publication = self.report.pubs[0]
        self.assertEqual(
            [x[2] for x in publication],
            [0, 25, 0, 0, 0, 0])

    def test_metric(self):
        self.assertEqual(self.report.metric, u"Book Section Requests")


class ParseLatin1(unittest.TestCase):
    """Tests for parsing BR2 in latin-1 encoding"""
    def setUp(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.report = pycounter.report.parse(
                os.path.join(os.path.dirname(__file__),
                             'data/simpleBR2_latin_1.csv'))

    def test_title(self):
        publication = self.report.pubs[1]
        self.assertEqual(publication.title, u'Öfake Book')



"""Test COUNTER JR1 journal report (CSV)"""

from __future__ import absolute_import

import datetime
import os
import unittest

import pytest

from pycounter import report


@pytest.mark.parametrize('pub_number,expected', [
    (0, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
    (1, [2, 1, 0, 0, 0, 5, 1, 1, 0, 5, 1, 0]),
])
def test_counter4_csv_data(csv_jr1_report_common_data, pub_number, expected):
    publication = csv_jr1_report_common_data.pubs[pub_number]
    assert [x[2] for x in publication] == expected


def test_metric(csv_jr1_report_std):
    assert csv_jr1_report_std.metric == u'FT Article Requests'


def test_customer(csv_jr1_report):
    assert csv_jr1_report.customer == u'University of Maximegalon'


def test_date_run(csv_jr1_report):
    assert csv_jr1_report.date_run == datetime.date(2012, 2, 21)


def test_period(csv_jr1_report):
    assert csv_jr1_report.period == (datetime.date(2011, 1, 1),
                                     datetime.date(2011, 12, 31))


def test_report_type(csv_jr1_report_std):
    assert csv_jr1_report_std.report_type == u'JR1'


def test_ibsn(csv_jr1_r4_report):
    assert csv_jr1_r4_report.pubs[1].isbn is None


class ParseExample(unittest.TestCase):
    """Tests for parsing C3 JR1"""
    def setUp(self):
        self.report = report.parse(os.path.join(os.path.dirname(__file__),
                                                'data/simpleJR1.csv'))

    def test_reportname(self):
        self.assertEqual(self.report.report_version, 3)

    def test_year(self):
        self.assertEqual(self.report.year, 2011)

    def test_platform(self):
        for publication in self.report:
            self.assertEqual(publication.publisher,
                             u"Maximegalon University Press")
            self.assertEqual(publication.platform, u"MJO")

    def test_html(self):
        expected = [0, 0]
        actual = [pub.html_total for pub in self.report.pubs]

        self.assertEqual(actual, expected)

    def test_pdf(self):
        expected = [0, 16]
        actual = [pub.pdf_total for pub in self.report.pubs]

        self.assertEqual(actual, expected)


class ParseMultiyear(unittest.TestCase):
    """Multi-year COUNTER report
    """

    def setUp(self):
        self.report = report.parse(os.path.join(os.path.dirname(__file__),
                                                'data/C4JR1my.csv'))

    def test_period(self):
        self.assertEqual(self.report.period,
                         (datetime.date(2011, 10, 1),
                          datetime.date(2012, 2, 29)))

    def test_monthdata_exception(self):
        self.assertRaises(AttributeError, getattr,
                          self.report.pubs[0], 'monthdata')

    def test_data(self):
        self.assertEqual(len(list(self.report.pubs[0])), 5)
        usage = [x[2] for x in self.report.pubs[0]]
        self.assertEqual(usage, [0, 0, 0, 0, 0])

    def test_month_data(self):
        expected = [datetime.date(2011, 10, 1),
                    datetime.date(2011, 11, 1),
                    datetime.date(2011, 12, 1),
                    datetime.date(2012, 1, 1),
                    datetime.date(2012, 2, 1),
                    ]

        months = [x[0] for x in self.report.pubs[0]]

        self.assertEqual(months, expected)


class ParseBigMultiyear(unittest.TestCase):
    """Multi-year report with more than 12 months of data
    """

    def setUp(self):
        self.report = report.parse(os.path.join(os.path.dirname(__file__),
                                                'data/C4JR1big.csv'))

    def test_period(self):
        self.assertEqual(self.report.period,
                         (datetime.date(2011, 1, 1),
                          datetime.date(2012, 12, 31)))

    def test_monthdata_exception(self):
        self.assertRaises(AttributeError, getattr,
                          self.report.pubs[0], 'monthdata')

    def test_data(self):
        usage = [x[2] for x in self.report.pubs[0]]
        self.assertEqual(usage,
                         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
                          ])


class ParseGOA(unittest.TestCase):
    """Gold Open Access Report
    """

    def setUp(self):
        self.report = report.parse(os.path.join(os.path.dirname(__file__),
                                                'data/C4JR1GOA.csv'))

    def test_metric(self):
        self.assertEqual(self.report.metric,
                         u"Gold Open Access Article Requests")


class ParseCounter4Bad(unittest.TestCase):
    """Tests for parsing C4 JR1 with questionable formatting..."""
    def setUp(self):
        self.report = report.parse(os.path.join(os.path.dirname(__file__),
                                                'data/C4JR1_bad.csv'))

    def test_counter4_csv_data(self):

        publication = self.report.pubs[0]
        self.assertEqual([x[2] for x in publication],
                         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        publication = self.report.pubs[1]
        self.assertEqual([x[2] for x in publication],
                         [2, 1, 0, 0, 0, 5, 1, 1, 0, 5, 1, 1000])



"""Test detecting file types correctly"""

from __future__ import absolute_import

import os
import unittest

from pycounter import report
from pycounter.test import common_data


class ParseTSV(common_data.TSVJR1):
    """Test TSV"""
    def setUp(self):
        self.report = report.parse(os.path.join(os.path.dirname(__file__),
                                                'data/tsvC4JR1'))


class ParseXLSX(unittest.TestCase):
    """Test XLSX"""
    def setUp(self):
        self.report = report.parse(os.path.join(os.path.dirname(__file__),
                                   'data/xlsxJR1'))

    def test_reportname(self):
        self.assertEqual(self.report.report_type, u'JR1')
        self.assertEqual(self.report.report_version, 4)

    def test_year(self):
        self.assertEqual(self.report.year, 2013)

    def test_platform(self):
        for publication in self.report:
            self.assertEqual(publication.publisher,
                             u"American Medical Association")
            self.assertEqual(publication.platform, u"Silverchair")

    def test_stats(self):
        publication = self.report.pubs[0]
        self.assertEqual([x[2] for x in publication],
                         [4, 14, 13, 30, 19, 7, 31, 6, 16, 12, 8, 12])
        publication = self.report.pubs[1]
        self.assertEqual(
            [x[2] for x in publication],
            [5414, 5459, 4936, 5172, 4064, 3904, 4054,
             4090, 5010, 6680, 5961, 3742])



"""Test COUNTER JR1 journal report (Excel)"""

from __future__ import absolute_import

import os

import pytest

from pycounter import report


@pytest.fixture(params=['JR1.xlsx', 'JR1_bad.xlsx'])
def jr1_report(request):
    return report.parse(os.path.join(os.path.dirname(__file__), 'data',
                                     request.param))


def test_report_type(jr1_report):
    assert jr1_report.report_type == u'JR1'


def test_report_version(jr1_report):
    assert jr1_report.report_version == 4


def test_year(jr1_report):
    assert jr1_report.year == 2013


def test_publisher(jr1_report):
    assert all(publication.publisher == u'American Medical Association'
               for publication in jr1_report)


def test_platform(jr1_report):
    assert all(publication.platform == u'Silverchair'
               for publication in jr1_report)


@pytest.mark.parametrize('pub_number,expected', [
    (0, [4, 14, 13, 30, 19, 7, 31, 6, 16, 12, 8, 12]),
    (1, [5414, 5459, 4936, 5172, 4064, 3904, 4054, 4090, 5010, 6680, 5961,
         3742]),
    (8, [592, 574, 502, 616, 349, 476, 460, 383, 434, 496, 522, 304]),
])
def test_stats(jr1_report, pub_number, expected):
    publication = jr1_report.pubs[pub_number]
    assert [x[2] for x in publication] == expected



"""Tests for pycounter.sushi"""
from __future__ import absolute_import

import datetime
import logging
import os
import unittest

from click.testing import CliRunner

from httmock import HTTMock, urlmatch
import mock

from pycounter import sushi
from pycounter import sushiclient
import pycounter.exceptions


@urlmatch(netloc=r'(.*\.)?example\.com$')
def sushi_mock(url_unused, request_unused):
    path = os.path.join(os.path.dirname(__file__),
                        'data', 'sushi_simple.xml')
    with open(path, 'rb') as datafile:
        return datafile.read().decode('utf-8')


@urlmatch(netloc=r'(.*\.)?example\.com$')
def error_mock(url_unused, request_unused):
    path = os.path.join(os.path.dirname(__file__),
                        'data', 'sushi_error.xml')
    with open(path, 'rb') as datafile:
        return datafile.read().decode('utf-8')


@urlmatch(netloc=r'(.*\.)?example\.com$')
def bogus_mock(url_unused, request_unused):
    return "Bogus response with no XML"


class TestHelpers(unittest.TestCase):
    """Test _ns helper"""

    def test_ns(self):
        self.assertEqual(
            sushi._ns("sushi", "name"),
            "{http://www.niso.org/schemas/sushi}name")


class TestConvertRawSimple(unittest.TestCase):
    """Test converting simple SUSHI response"""

    def setUp(self):
        path = os.path.join(os.path.dirname(__file__),
                            'data', 'sushi_simple.xml')
        with open(path, 'rb') as datafile:
            self.report = sushi._raw_to_full(datafile.read())

    def test_report(self):
        self.assertEqual(self.report.report_type, u'JR1')
        self.assertEqual(self.report.report_version, 4)

    def test_customer(self):
        self.assertEqual(self.report.institutional_identifier,
                         u"exampleLibrary")

    def test_title(self):
        publication = next(iter(self.report))
        self.assertEqual(publication.title, u'Journal of fake data')

    def test_data(self):
        publication = next(iter(self.report))
        self.assertEqual(publication.html_total, 6)
        self.assertEqual(publication.pdf_total, 8)
        self.assertEqual(publication.doi, u'10.5555/12345678')
        self.assertEqual(publication.proprietary_id, u'JFD')
        data = [month[2] for month in publication]
        self.assertEqual(data[0], 14)


class TestConvertRawBook(unittest.TestCase):
    """Test converting simple BR1 SUSHI response"""

    def setUp(self):
        path = os.path.join(os.path.dirname(__file__),
                            'data', 'sushi_simple_br1.xml')
        with open(path, 'rb') as datafile:
            self.report = sushi._raw_to_full(datafile.read())

    def test_report(self):
        self.assertEqual(self.report.report_type, u'BR1')
        self.assertEqual(self.report.report_version, 4)

    def test_customer(self):
        self.assertEqual(self.report.institutional_identifier,
                         u"exampleLibrary")

    def test_isbn(self):
        i_report = iter(self.report)
        publication = next(i_report)
        self.assertEqual(publication.isbn, u"9780011234549")
        next(i_report)
        pub3 = next(i_report)
        self.assertEqual(pub3.isbn, u"9780011234540")

    def test_title(self):
        publication = next(iter(self.report))
        self.assertEqual(publication.title, u'Fake data')

    def test_data(self):
        publication = next(iter(self.report))
        data = [month[2] for month in publication]
        self.assertEqual(data[0], 14)

    def test_proprietary(self):
        publication = next(iter(self.report))
        self.assertEqual(publication.proprietary_id, u'FD')


class TestConvertRawDatabase(unittest.TestCase):
    """Test converting simple DB1 SUSHI response"""

    def setUp(self):
        path = os.path.join(os.path.dirname(__file__),
                            'data', 'sushi_simple_db1.xml')
        with open(path, 'rb') as datafile:
            self.report = sushi._raw_to_full(datafile.read())
        self.databases = list(self.report)

    def test_report(self):
        self.assertEqual(self.report.report_type, u'DB1')
        self.assertEqual(self.report.report_version, 4)

    def test_customer(self):
        self.assertEqual(self.report.institutional_identifier,
                         u"exampleLibrary")

    def test_platform(self):
        database = self.databases[0]
        self.assertEqual(database.platform, u'ExamplePlatform')

    def test_publisher(self):
        database = self.databases[0]
        self.assertEqual(database.publisher, u'ExamplePublisher')

    def test_title(self):
        database = self.databases[0]
        self.assertEqual(database.title, u'ExampleDatabase')

    def test_search_reg(self):
        database = self.databases[0]
        data = [month[2] for month in database]
        self.assertEqual(database.metric, u'Regular Searches')
        self.assertEqual(data[0], 5)

    def test_search_fed(self):
        database = self.databases[1]
        data = [month[2] for month in database]
        self.assertEqual(database.metric, u'Searches-federated and automated')
        self.assertEqual(data[0], 13)

    def test_result_click(self):
        database = self.databases[2]
        data = [month[2] for month in database]
        self.assertEqual(database.metric, u'Result Clicks')
        self.assertEqual(data[0], 16)

    def test_record_view(self):
        database = self.databases[3]
        data = [month[2] for month in database]
        self.assertEqual(database.metric, u'Record Views')
        self.assertEqual(data[0], 7)


class TestRawDatabaseWithMissingData(unittest.TestCase):
    """
    Test database request with January missing for 'search_fed'
    and no 'record_view' records
    """

    def setUp(self):
        path = os.path.join(os.path.dirname(__file__),
                            'data', 'sushi_db1_missing_record_view.xml')
        with open(path, 'rb') as datafile:
            self.report = sushi._raw_to_full(datafile.read())
        # missing data only injected when making generic to write
        self.report.as_generic()
        self.databases = list(self.report)

    def test_january(self):
        database = self.databases[1]
        data = [month[2] for month in database]
        self.assertEqual(database.metric, u'Searches-federated and automated')
        self.assertEqual(data[0], 0)

    def test_record_view(self):
        database = self.databases[3]
        data = [month[2] for month in database]
        self.assertEqual(database.metric, u'Record Views')
        self.assertEqual(data[0], 0)


class TestMissingMonth(unittest.TestCase):
    """Test SUSHI with months missing"""

    def setUp(self):
        path = os.path.join(os.path.dirname(__file__),
                            'data', 'sushi_missing_jan.xml')
        with open(path, 'rb') as datafile:
            self.report = sushi._raw_to_full(datafile.read())
        self.publication = next(iter(self.report))

    def test_february(self):
        first_month_data = next(iter(self.publication))
        self.assertEqual(first_month_data[0],
                         datetime.date(2013, 2, 1))

    def test_title(self):
        self.assertEqual(self.publication.title, u'Journal of fake data')


class TestSushiRequest(unittest.TestCase):
    """Test making a SUSHI request"""

    def setUp(self):
        with HTTMock(sushi_mock):
            self.report = sushi.get_report('http://www.example.com/Sushi',
                                           datetime.date(2015, 1, 1),
                                           datetime.date(2015, 1, 31)
                                           )

    def test_report(self):
        self.assertEqual(self.report.report_type, u'JR1')
        self.assertEqual(self.report.report_version, 4)


class TestSushiDump(unittest.TestCase):
    """Test dumping SUSHI response"""

    @mock.patch('pycounter.sushi.logger')
    def test_dump(self, mock_logger):
        with HTTMock(sushi_mock):
            self.report = sushi.get_report('http://www.example.com/Sushi',
                                           datetime.date(2015, 1, 1),
                                           datetime.date(2015, 1, 31),
                                           sushi_dump=True
                                           )
        self.assertTrue(mock_logger.debug.called)


class TestBogusXML(unittest.TestCase):
    """Test dealing with broken XML"""

    def test_request(self):
        logging.disable(logging.CRITICAL)
        with HTTMock(bogus_mock):
            self.assertRaises(pycounter.exceptions.SushiException,
                              sushi.get_report,
                              'http://www.example.com/bogus',
                              datetime.date(2015, 1, 1),
                              datetime.date(2015, 1, 31)
                              )


class TestSushiError(unittest.TestCase):
    """Test error from SUSHI"""

    def test_error(self):
        with HTTMock(error_mock):
            self.assertRaises(pycounter.exceptions.SushiException,
                              sushi.get_report,
                              'http://www.example.com/out_of_range',
                              datetime.date(2015, 1, 1),
                              datetime.date(2015, 1, 31)
                              )


class TestSushiClient(unittest.TestCase):
    """Test the client"""

    def test_get(self):
        arglist = [
            'http://www.example.com/Sushi',
        ]

        with HTTMock(sushi_mock):
            runner = CliRunner()
            with runner.isolated_filesystem():
                result = runner.invoke(sushiclient.main, arglist)
                with open('report.tsv') as tsv_file:
                    self.assertTrue('Journal Report 1' in tsv_file.read())
                self.assertEqual(result.exit_code, 0)

    def test_end_date_error(self):
        """Test trying to use implied start date and explicit end date"""
        arglist = [
            'http://www.example.com/Sushi',
            '-e', '2015-12-31',
        ]
        runner = CliRunner()
        result = runner.invoke(sushiclient.main, arglist)
        self.assertEqual(result.exit_code, 1)

    def test_explicit_dates(self):
        """Test providing both start and end dates"""
        arglist = [
            'http://www.example.com/Sushi',
            '-s', '2015-10-31',
            '-e', '2015-12-31',
        ]
        with HTTMock(sushi_mock):
            runner = CliRunner()
            with runner.isolated_filesystem():
                result = runner.invoke(sushiclient.main, arglist)
                self.assertEqual(result.exit_code, 0)


class TestMissingItemIdentifier(unittest.TestCase):
    """Test converting simple SUSHI response"""

    def setUp(self):
        path = os.path.join(os.path.dirname(__file__),
                            'data', 'sushi_missing_ii.xml')
        with open(path, 'rb') as datafile:
            self.report = sushi._raw_to_full(datafile.read())

    def test_issn(self):
        publication = next(iter(self.report))
        self.assertEqual(publication.issn, "")



"""Test COUNTER JR1 journal report (TSV)"""

from __future__ import absolute_import

import os

from pycounter import report
from pycounter.test import common_data


class ParseExample(common_data.TSVJR1):
    """Test tab-separated JR1"""

    def setUp(self):
        self.report = report.parse(os.path.join(os.path.dirname(__file__),
                                                'data/simpleJR1.tsv'))

    def test_html(self):
        expected = [0, 15, 33, 0]
        actual = [pub.html_total for pub in self.report.pubs]

        self.assertEqual(actual, expected)

    def test_pdf(self):
        expected = [32, 12, 855, 40]
        actual = [pub.pdf_total for pub in self.report.pubs]

        self.assertEqual(actual, expected)



"""Tests for the helpers module"""

from __future__ import absolute_import

import datetime

import pytest

from pycounter.helpers import next_month, prev_month


@pytest.mark.parametrize('pair', [((2000, 1, 1), (2000, 2, 1)),
                                  ((2000, 12, 1), (2001, 1, 1)),
                                  ((2000, 2, 29), (2000, 3, 1)),
                                  ((1999, 12, 6), (2000, 1, 1)),
                                  ])
def test_nextmonth(pair):
    assert datetime.date(*pair[1]) == next_month(datetime.date(*pair[0]))


@pytest.mark.parametrize('pair', [((2000, 1, 1), (1999, 12, 1)),
                                  ((2000, 12, 1), (2000, 11, 1)),
                                  ((2000, 2, 29), (2000, 1, 1)),
                                  ((1999, 12, 6), (1999, 11, 1)),
                                  ])
def test_prevmonth(pair):
    assert datetime.date(*pair[1]) == prev_month(datetime.date(*pair[0]))



"""Test parsing of COUNTER BR1 book report."""

from __future__ import absolute_import

import datetime
import os
import unittest

import pycounter.report


class ParseExample(unittest.TestCase):
    """Tests for parsing C3 BR1"""
    def setUp(self):
        self.report = pycounter.report.parse(
            os.path.join(os.path.dirname(__file__),
                         'data/simpleBR1.csv'))

    def test_reportname(self):
        self.assertEqual(self.report.report_type, u'BR1')
        self.assertEqual(self.report.report_version, 1)

    def test_year(self):
        self.assertEqual(self.report.year, 2012)

    def test_platform(self):
        for publication in self.report:
            self.assertEqual(publication.publisher, u"Megadodo Publications")
            self.assertEqual(publication.platform, u"HHGTTG Online")

    def test_stats(self):
        publication = self.report.pubs[0]
        self.assertEqual(
            [x[2] for x in publication],
            [0, 25, 0, 0, 0, 0])

    def test_customer(self):
        self.assertEqual(self.report.customer,
                         u"University of Maximegalon")

    def test_date_run(self):
        self.assertEqual(self.report.date_run, datetime.date(2012, 7, 9))

    def test_period(self):
        self.assertEqual(self.report.period,
                         (datetime.date(2012, 1, 1),
                          datetime.date(2012, 6, 30)))


class ParseCounter4Example(unittest.TestCase):
    """Tests for parsing C4 BR1"""
    def setUp(self):
        self.report = pycounter.report.parse(
            os.path.join(os.path.dirname(__file__),
                         'data/C4BR1.tsv'))

    def test_reportname(self):
        self.assertEqual(self.report.report_type, u'BR1')
        self.assertEqual(self.report.report_version, 4)

    def test_year(self):
        self.assertEqual(self.report.year, 2012)

    def test_platform(self):
        for publication in self.report:
            self.assertEqual(publication.publisher, u"Megadodo Publications")
            self.assertEqual(publication.platform, u"HHGTTG Online")

    def test_stats(self):
        publication = self.report.pubs[0]
        self.assertEqual(
            [x[2] for x in publication],
            [0, 25, 0, 0, 0, 0])

    def test_metric(self):
        self.assertEqual(self.report.metric, u"Book Title Requests")

    def test_customer(self):
        self.assertEqual(self.report.customer,
                         u"University of Maximegalon")

    def test_date_run(self):
        self.assertEqual(self.report.date_run, datetime.date(2012, 7, 9))

    def test_period(self):
        self.assertEqual(self.report.period,
                         (datetime.date(2012, 1, 1),
                          datetime.date(2012, 6, 30)))

    def test_isbn(self):
        publication = self.report.pubs[0]
        self.assertEqual(publication.isbn, u'9787490833809')



"""command line client to fetch statistics via SUSHI."""
from __future__ import absolute_import
from __future__ import print_function

import datetime
import logging
import sys

import click

from pycounter import sushi
from pycounter.helpers import convert_date_run, last_day, prev_month

logging.basicConfig()


@click.command()
@click.argument('url')
@click.option('--report', '-r', default='JR1',
              help='report name (default JR1)')
@click.option('--release', '-l', default=4,
              help='COUNTER release (default 4)')
@click.option('--start_date', '-s',
              help='Start Date YYYY-MM-DD (default first day of last month)')
@click.option('--end_date', '-e',
              help='End Date YYYY-MM-DD (default last day of last month OR '
                   'last day of start month')
@click.option('--requestor_id', '-i',
              help='Requestor ID')
@click.option('--requestor_email',
              help='Email address of requestor')
@click.option('--requestor_name',
              help='Internationally recognized organization name')
@click.option('--customer_reference', '-c',
              help='Customer reference')
@click.option('--customer_name',
              help='Internationally recognized organization name')
@click.option('--format', '-f', 'format_', default='tsv',
              help='Output format (default tsv)',
              type=click.Choice(['tsv']))
@click.option('--output_file', '-o', default='report.%s',
              help='Output file to write (will be overwritten)',
              type=click.Path(writable=True))
@click.option('--dump', '-d', is_flag=True)
def main(url, report, release, start_date, end_date, requestor_id,
         requestor_email, requestor_name, customer_name,
         customer_reference, format_, output_file, dump):
    """Main function for the SUSHI client."""
    click.echo("pycounter SUSHI client for URL %s (%s R%s)"
               % (url, report, release))
    if end_date is not None and start_date is None:
        click.echo('Cannot specify --end_date without --start_date',
                   err=True)
        sys.exit(1)
    if start_date is None:
        converted_start_date = prev_month(datetime.datetime.now())
    else:
        converted_start_date = convert_date_run(start_date)
    if end_date is None:
        converted_end_date = last_day(converted_start_date)
    else:
        converted_end_date = convert_date_run(end_date)
    report = sushi.get_report(wsdl_url=url,
                              report=report,
                              release=release,
                              requestor_id=requestor_id,
                              requestor_name=requestor_name,
                              requestor_email=requestor_email,
                              customer_reference=customer_reference,
                              customer_name=customer_name,
                              start_date=converted_start_date,
                              end_date=converted_end_date,
                              sushi_dump=dump)
    output_file = output_file % format_
    report.write_to_file(output_file, format_)


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



"""Read CSV as unicode from both python 2 and 3 transparently."""
from __future__ import absolute_import

import csv
import warnings

import six


# noinspection PyUnusedLocal
class UnicodeReader(six.Iterator):
    """CSV reader that can handle unicode.

    Must be used as a context manager:

    with UnicodeReader('myfile.csv') as reader:
        pass # do things with reader

    :param filename: path to file to open
    :param dialect: a csv.Dialect instance or dialect name
    :param encoding: text encoding of file
    :param fallback_encoding: encoding to fall back to if default
             encoding fails; gives warning if it's used.

    All other parameters will be passed through to csv.reader()
    """

    def __init__(self, filename, dialect=csv.excel,
                 encoding="utf-8", fallback_encoding="latin-1", **kwargs):
        self.filename = filename
        self.dialect = dialect
        self.encoding = encoding
        self.kwargs = kwargs
        self.fileobj = None
        self.reader = None
        self.fallback_encoding = fallback_encoding

    def __enter__(self):
        if six.PY3:
            self.fileobj = open(self.filename, 'rt',
                                encoding=self.encoding, newline='')
            try:
                self.fileobj.read()
            except UnicodeDecodeError:
                warnings.warn("Decoding with '%s' codec failed; falling "
                              "back to '%s'" % (self.encoding,
                                                self.fallback_encoding))
                self.fileobj = open(self.filename, 'rt',
                                    encoding=self.fallback_encoding,
                                    newline='')
                self.encoding = self.fallback_encoding
            finally:
                self.fileobj.seek(0)
        else:
            self.fileobj = open(self.filename, 'rb')
            try:
                self.fileobj.read().decode(self.encoding)
            except UnicodeDecodeError:
                warnings.warn("Decoding with '%s' codec failed; falling "
                              "back to '%s'" % (self.encoding,
                                                self.fallback_encoding))
                self.encoding = self.fallback_encoding
            finally:
                self.fileobj.seek(0)
        self.reader = csv.reader(self.fileobj, dialect=self.dialect,
                                 **self.kwargs)
        return self

    def __exit__(self, type_, value, traceback):
        self.fileobj.close()

    def __next__(self):
        row = next(self.reader)
        if six.PY3:
            return row
        return [s.decode(self.encoding) for s in row]

    def __iter__(self):
        return self


# noinspection PyUnusedLocal
class UnicodeWriter(object):
    """CSV writer that can handle unicode.

    Must be used as a context manager:

    with UnicodeWriter('myfile.csv') as writer:
        pass # do things with writer

    :param filename: path to file to open
    :param dialect: a csv.Dialect instance or dialect name
    :param encoding: text encoding of file

    All other parameters will be passed through to csv.writer()
    """

    def __init__(self, filename, dialect=csv.excel,
                 encoding="utf-8", lineterminator='\n', **kwargs):
        self.filename = filename
        self.dialect = dialect
        self.encoding = encoding
        self.lineterminator = lineterminator
        self.kwargs = kwargs
        self.writer = None
        self.fileobj = None

    def __enter__(self):
        if six.PY3:
            self.fileobj = open(self.filename, 'wt',
                                encoding=self.encoding, newline='')
        else:
            self.fileobj = open(self.filename, 'wb')
        self.writer = csv.writer(self.fileobj, dialect=self.dialect,
                                 lineterminator=self.lineterminator,
                                 **self.kwargs)
        return self

    def __exit__(self, type_, value, traceback):
        self.fileobj.close()

    def writerow(self, row):
        """Write a row to the output.

        :param row: list of cells to write to the file
        """
        if not six.PY3:
            row = [(s or '').encode(self.encoding) for s in row]
        self.writer.writerow(row)

    def writerows(self, rows):
        """Write many rows to the output.

        :param rows: list of lists of cells to write
        """
        for row in rows:
            self.writerow(row)



# -*- coding: utf-8 -*-
#
# pycounter documentation build configuration file, created by
# sphinx-quickstart on Thu Nov 13 08:44:21 2014.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('../../pycounter'))
sys.path.append(os.path.join(os.path.dirname(__file__), '../..'))

# noinspection PyProtectedMember
from pycounter.version import __version__

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.

# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.coverage',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.

# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'pycounter'
copyright = u'2014-2016, University of Pittsburgh'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = __version__
# The full version, including alpha/beta/rc tags.
release = version

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = 'pycounterdoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {

}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    ('index', 'pycounter.tex', u'pycounter Documentation',
     u'Geoffrey Spear', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'pycounter', u'pycounter Documentation',
     [u'Geoffrey Spear'], 1)
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    ('index', 'pycounter', u'pycounter Documentation',
     u'Geoffrey Spear', 'pycounter', 'One line description of project.',
     'Miscellaneous'),
]



import os

import pytest

from pycounter import report


@pytest.fixture(params=['csvC4JR1', 'C4JR1.csv', 'simpleJR1.csv',
                        'C4JR1_bad.csv', 'C4JR1GOA.csv'])
def csv_jr1_report(request):
    return report.parse(os.path.join(os.path.dirname(__file__),
                                     'data', request.param))


@pytest.fixture(params=['csvC4JR1', 'C4JR1.csv', 'simpleJR1.csv',
                        'C4JR1_bad.csv'])
def csv_jr1_report_std(request):
    return report.parse(os.path.join(os.path.dirname(__file__),
                                     'data', request.param))


@pytest.fixture(params=['csvC4JR1', 'C4JR1.csv', 'simpleJR1.csv'])
def csv_jr1_report_common_data(request):
    return report.parse(os.path.join(os.path.dirname(__file__),
                                     'data', request.param))


@pytest.fixture(params=['csvC4JR1', 'C4JR1.csv', 'C4JR1_bad.csv'])
def csv_jr1_r4_report(request):
    return report.parse(os.path.join(os.path.dirname(__file__),
                                     'data', request.param))



"""Test parsing of deliberately bad data."""

from __future__ import absolute_import

import unittest

from pycounter import report
import pycounter.exceptions


class TestParseBad(unittest.TestCase):
    def test_bogus_report_type(self):
        data = [[u"Bogus Report 7 (R4)"]]
        self.assertRaises(pycounter.exceptions.UnknownReportTypeError,
                          report.parse_generic,
                          iter(data)
                          )

    def test_unsupported_report_type(self):
        """Test that we fail for report types that are real but unsupported"""
        # FIXME: eventually should be supported; remove this test when all are
        data = [[u"Platform Report 1 (R4)"]]
        self.assertRaises(pycounter.exceptions.UnknownReportTypeError,
                          report.parse_generic,
                          iter(data)
                          )


class TestBogusFiletype(unittest.TestCase):
    def test_bogus_file_type(self):
        self.assertRaises(pycounter.exceptions.PycounterException,
                          report.parse,
                          'no_such_file',
                          'qsx'
                          )



"""NISO SUSHI support."""
from __future__ import absolute_import

import collections
import datetime
import logging
import uuid


import arrow

from lxml import etree
from lxml import objectify

import requests
import six

import pycounter.constants
import pycounter.exceptions
from pycounter.helpers import convert_date_run
import pycounter.report

logger = logging.getLogger(__name__)
NS = pycounter.constants.NS


def get_sushi_stats_raw(wsdl_url, start_date, end_date, requestor_id=None,
                        requestor_email=None, requestor_name=None,
                        customer_reference=None, customer_name=None,
                        report="JR1", release=4, sushi_dump=False):
    """Get SUSHI stats for a given site in raw XML format.

    :param wsdl_url: URL to SOAP WSDL for this provider

    :param start_date: start date for report (must be first day of a month)

    :param end_date: end date for report (must be last day of a month)

    :param requestor_id: requestor ID as defined by SUSHI protocol

    :param requestor_email: requestor email address, if required by provider

    :param requestor_name: Internationally recognized organization name

    :param customer_reference: customer reference number as defined by SUSHI
        protocol

    :param customer_name: Internationally recognized organization name

    :param report: report type, values defined by SUSHI protocol

    :param release: report release number (should generally be `4`.)

    :param sushi_dump: produces dump of XML to DEBUG logger

    """
    root = etree.Element("{%(SOAP-ENV)s}Envelope" % NS, nsmap=NS)
    body = etree.SubElement(root, "{%(SOAP-ENV)s}Body" % NS)
    timestamp = arrow.utcnow().isoformat()
    rr = etree.SubElement(body, "{%(sushicounter)s}ReportRequest" % NS,
                          {
                              'Created': timestamp,
                              'ID': str(uuid.uuid4())
                          })

    req = etree.SubElement(rr, "{%(sushi)s}Requestor" % NS)
    rid = etree.SubElement(req, "{%(sushi)s}ID" % NS)
    rid.text = requestor_id
    req_name_element = etree.SubElement(req, "{%(sushi)s}Name" % NS)
    req_name_element.text = requestor_name
    req_email_element = etree.SubElement(req, "{%(sushi)s}Email" % NS)
    req_email_element.text = requestor_email

    cust_ref_elem = etree.SubElement(rr, "{%(sushi)s}CustomerReference" % NS)
    cid = etree.SubElement(cust_ref_elem, "{%(sushi)s}ID" % NS)
    cid.text = customer_reference
    cust_name_elem = etree.SubElement(cust_ref_elem, "{%(sushi)s}Name" % NS)
    cust_name_elem.text = customer_name

    report_def_elem = etree.SubElement(rr, "{%(sushi)s}ReportDefinition" % NS,
                                       Name=report, Release=str(release))
    filters = etree.SubElement(report_def_elem, "{%(sushi)s}Filters" % NS)
    udr = etree.SubElement(filters, "{%(sushi)s}UsageDateRange" % NS)
    beg = etree.SubElement(udr, "{%(sushi)s}Begin" % NS)
    beg.text = start_date.strftime("%Y-%m-%d")
    end = etree.SubElement(udr, "{%(sushi)s}End" % NS)
    end.text = end_date.strftime("%Y-%m-%d")

    payload = etree.tostring(root, pretty_print=True,
                             xml_declaration=True, encoding="utf-8")

    headers = {"SOAPAction": '"SushiService:GetReportIn"',
               "Content-Type": "text/xml; charset=UTF-8",
               "User-Agent": "pycounter/%s" % pycounter.__version__,
               "Content-Length": str(len(payload))}

    response = requests.post(url=wsdl_url,
                             headers=headers,
                             data=payload,
                             verify=False)

    if sushi_dump:
        logger.debug("SUSHI DUMP: request: %s \n\n response: %s",
                     payload,
                     response.content)
    return response.content


def get_report(*args, **kwargs):
    """Get a usage report from a SUSHI server.

    returns a :class:`pycounter.report.CounterReport` object.

    parameters: see get_sushi_stats_raw
    """
    raw_report = get_sushi_stats_raw(*args, **kwargs)
    return _raw_to_full(raw_report)


def _ns(namespace, name):
    """Convenience function to make a namespaced XML name.

    :param namespace: one of 'SOAP-ENV', 'sushi', 'sushicounter', 'counter'
    :param name: tag name within the given namespace
    """
    return "{" + NS[namespace] + "}" + name


def _raw_to_full(raw_report):
    """Convert a raw report to CounterReport.

    :param raw_report: raw XML report
    :return: a :class:`pycounter.report.CounterReport`
    """
    try:
        root = etree.fromstring(raw_report)
    except etree.XMLSyntaxError:
        logger.error("XML syntax error: %s", raw_report)
        raise pycounter.exceptions.SushiException(
            message="XML syntax error",
            raw=raw_report)
    o_root = objectify.fromstring(raw_report)
    rep = None
    try:
        rep = o_root.Body[_ns('sushicounter', "ReportResponse")]
        c_report = rep.Report[_ns('counter', 'Report')]
    except AttributeError:
        try:
            c_report = rep.Report[_ns('counter', 'Reports')].Report
        except AttributeError:
            logger.error("report not found in XML: %s", raw_report)
            raise pycounter.exceptions.SushiException(
                message="report not found in XML",
                raw=raw_report, xml=o_root)
    logger.debug("COUNTER report: %s", etree.tostring(c_report))
    start_date = datetime.datetime.strptime(
        root.find('.//%s' % _ns('sushi', 'Begin')).text,
        "%Y-%m-%d").date()

    end_date = datetime.datetime.strptime(
        root.find('.//%s' % _ns('sushi', 'End')).text,
        "%Y-%m-%d").date()

    report_data = {'period': (start_date, end_date)}

    rep_def = root.find('.//%s' % _ns('sushi', 'ReportDefinition'))
    report_data['report_version'] = int(rep_def.get('Release'))

    report_data['report_type'] = rep_def.get('Name')

    customer = root.find('.//%s' % _ns('counter', 'Customer'))
    try:
        report_data['customer'] = (customer.find('.//%s' %
                                                 _ns('counter', 'Name')).text)
    except AttributeError:
        report_data['customer'] = ""

    inst_id = customer.find('.//%s' % _ns('counter', 'ID')).text
    report_data['institutional_identifier'] = inst_id

    rep_root = root.find('.//%s' % _ns('counter', 'Report'))
    created_string = rep_root.get('Created')
    if created_string is not None:
        report_data['date_run'] = arrow.get(created_string)
    else:
        report_data['date_run'] = datetime.datetime.now()

    report = pycounter.report.CounterReport(**report_data)

    report.metric = pycounter.constants.METRICS.get(report_data['report_type'])

    for item in c_report.Customer.ReportItems:
        try:
            publisher_name = item.ItemPublisher.text
        except AttributeError:
            publisher_name = ""
        title = item.ItemName.text
        platform = item.ItemPlatform.text

        eissn = issn = ""
        print_isbn = None
        online_isbn = None
        doi = ""
        prop_id = ""

        try:
            for identifier in item.ItemIdentifier:
                if identifier.Type == "Print_ISSN":
                    issn = identifier.Value.text
                    if issn is None:
                        issn = ""
                elif identifier.Type == "Online_ISSN":
                    eissn = identifier.Value.text
                    if eissn is None:
                        eissn = ""
                elif identifier.Type == "Online_ISBN":
                    online_isbn = identifier.Value.text
                elif identifier.Type == "Print_ISBN":
                    print_isbn = identifier.Value.text
                elif identifier.Type == "DOI":
                    doi = identifier.Value.text
                elif identifier.Type == "Proprietary":
                    prop_id = identifier.Value.text

        except AttributeError:
            pass

        month_data = []
        html_usage = 0
        pdf_usage = 0

        metrics_for_db = collections.defaultdict(list)

        for perform_item in item.ItemPerformance:
            item_date = convert_date_run(perform_item.Period.Begin.text)
            logger.debug("perform_item date: %r", item_date)
            usage = None
            if hasattr(perform_item, 'Instance'):
                for inst in perform_item.Instance:
                    if inst.MetricType == "ft_total":
                        usage = str(inst.Count)
                    elif inst.MetricType == "ft_pdf":
                        pdf_usage += int(inst.Count)
                    elif inst.MetricType == "ft_html":
                        html_usage += int(inst.Count)
                    elif report.report_type.startswith('DB'):
                        metrics_for_db[inst.MetricType].append(
                            (item_date, int(inst.Count)))
            if usage is not None:
                month_data.append((item_date, int(usage)))

        if report.report_type:
            if report.report_type.startswith('JR'):
                report.pubs.append(pycounter.report.CounterJournal(
                    title=title,
                    platform=platform,
                    publisher=publisher_name,
                    period=report.period,
                    metric=report.metric,
                    issn=issn,
                    eissn=eissn,
                    doi=doi,
                    proprietary_id=prop_id,
                    month_data=month_data,
                    html_total=html_usage,
                    pdf_total=pdf_usage
                ))
            elif report.report_type.startswith('BR'):
                report.pubs.append(
                    pycounter.report.CounterBook(
                        title=title,
                        platform=platform,
                        publisher=publisher_name,
                        period=report.period,
                        metric=report.metric,
                        issn=issn,
                        doi=doi,
                        proprietary_id=prop_id,
                        print_isbn=print_isbn,
                        online_isbn=online_isbn,
                        month_data=month_data,
                    ))
            elif report.report_type.startswith('DB'):
                for metric_code, month_data in six.iteritems(metrics_for_db):
                    metric = pycounter.constants.DB_METRIC_MAP[metric_code]
                    report.pubs.append(
                        pycounter.report.CounterDatabase(
                            title=title,
                            platform=platform,
                            publisher=publisher_name,
                            period=report.period,
                            metric=metric,
                            month_data=month_data
                        ))

    return report



"""Exception classes for pycounter."""


class PycounterException(Exception):
    """Base class for all module exceptions."""

    pass


class UnknownReportTypeError(PycounterException):
    """We can't parse this kind of report yet."""

    pass


class SushiException(PycounterException):
    """Base class for SUSHI-related exceptions."""

    def __init__(self, message, raw=None, xml=None):
        super(SushiException, self).__init__(message)
        self.raw = raw
        self.xml = xml


class ServiceNotAvailableError(SushiException):
    """Fatal error: service failed due to internal error."""

    pass


class ServiceBusyError(SushiException):
    """Fatal error: server is too busy; try again later."""

    pass


class TooManyRequestsError(SushiException):
    """Fatal error: The client has made too many requests to the service."""

    pass


class RequestorNotAuthorizedError(SushiException):
    """
    Requestor is not authorized.

    Requestor ID or CustomerReference are not recognized as giving
    authorization to retrieve statistics.
    """

    pass


class ReportNotSupportedError(SushiException):
    """Server cannot serve the requested report name or version."""

    pass


class InvalidDateError(SushiException):
    """Dates are formatted incorrectly or illogical."""

    pass


class NoUsageAvailableError(SushiException):
    """Service has no data for requested date range."""

    pass


class SushiWarning(Warning, SushiException):
    """Base class for SUSHI_related warnings."""

    pass


class PartialDataWarning(SushiWarning):
    """Request not completely fulfilled, but available data was returned."""

    pass


class FilterNotSupportedWarning(SushiWarning):
    """a filter was ignored, because it was unsupported.

    Data is returned without that filter applied
    """

    pass


class ReportAttributeNotSupportedWarning(SushiWarning):
    """a report attribute was ignored because it was unsupported.

    Data is returned without that attribute applied
    """

    pass


class InvalidFilterValueWarning(SushiWarning):
    """a filter was ignored, because its value was unsupported.

    Data is returned without that filter applied
    """

    pass


class InvalidReportAttributeWarning(SushiWarning):
    """a report attribute was ignored because its value was unsupported.

    Data is returned without that attribute applied
    """

    pass



"""COUNTER journal and book reports and associated functions."""

from __future__ import absolute_import

import collections
import datetime
import logging
import re
import warnings

import arrow
import six

from pycounter import csvhelper
from pycounter.constants import CODES, HEADER_FIELDS, METRICS
from pycounter.constants import REPORT_DESCRIPTIONS, TOTAL_TEXT
from pycounter.exceptions import PycounterException, UnknownReportTypeError
from pycounter.helpers import convert_covered, convert_date_column, \
    convert_date_run, format_stat, guess_type_from_content, last_day, \
    next_month


class CounterReport(object):
    """
    a COUNTER usage statistics report.

    Iterate over the report object to get its rows (each of which is a
    :class:`CounterBook <CounterBook>` or :class:`CounterJournal
    <CounterJournal>` instance.

    :param metric: metric being tracked by this report. For database
        reports (which have multiple metrics per report), this should be
        set to `None`.

    :param report_type: type of report (e.g., "JR1", "BR2")

    :param report_version: COUNTER version

    :param customer: name of customer on report

    :param institutional_identifier: unique ID assigned by vendor for
        customer

    :param period: tuple of datetime.date objects corresponding to the
        beginning and end of the covered range

    :param date_run: date the COUNTER report was generated

    :param section_type: predominant section type used for this report.
        (applies to report BR2; should probably be None for any other report
        type)

    """

    def __init__(self, report_type=None, report_version=4, metric=None,
                 customer=None, institutional_identifier=None,
                 period=(None, None), date_run=None,
                 section_type=None):
        self.pubs = []
        self.report_type = report_type
        self.report_version = report_version
        self.metric = metric
        self.customer = customer
        self.institutional_identifier = institutional_identifier
        self.period = period
        if date_run is None:
            self.date_run = datetime.date.today()
        else:
            self.date_run = date_run
        self.year = None
        self.section_type = section_type

    def __repr__(self):
        return (
            "<CounterReport %s version %s for date range %s to %s>" %
            (self.report_type,
             self.report_version,
             self.period[0], self.period[1])
        )

    def __iter__(self):
        return iter(self.pubs)

    def write_to_file(self, path, format_):
        """
        Output report to a file.

        :param path: location to write file
        :param format_: file format. Currently supports 'tsv'
        :return:
        """
        if format_ == 'tsv':
            self.write_tsv(path)
        else:
            raise PycounterException("unknown file type %s" % format_)

    def write_tsv(self, path):
        """
        Output report to a COUNTER 4 TSV file.

        :param path: location to write file
        """
        lines = self.as_generic()
        with csvhelper.UnicodeWriter(path, delimiter='\t') as writer:
            writer.writerows(lines)

    def as_generic(self):
        """
        Output report as list of lists.

        Nested list will contain cells that would appear
        in COUNTER report (suitable for writing as CSV, TSV, etc.)
        """
        output_lines = []
        rep_type = ""
        for name, code in CODES.items():
            if code == self.report_type[0:2]:
                rep_type = name

        report_name = ("%s Report %s (R%s)" %
                       (rep_type, self.report_type[-1], self.report_version))
        output_lines.append([report_name,
                             REPORT_DESCRIPTIONS[self.report_type]])
        if self.report_type == 'BR2':
            output_lines.append([self.customer, u'Section Type:'])
            output_lines.append([self.institutional_identifier,
                                 self.section_type])
        else:
            output_lines.append([self.customer])
            output_lines.append([self.institutional_identifier])
        output_lines.append([u'Period covered by Report:'])
        period = "%s to %s" % (
            self.period[0].strftime('%Y-%m-%d'),
            self.period[1].strftime('%Y-%m-%d')
        )
        output_lines.append([period])
        output_lines.append([u'Date run:'])
        output_lines.append([self.date_run.strftime('%Y-%m-%d')])
        output_lines.append(self._table_header())
        if self.report_type in ('JR1', 'BR1', 'BR2', 'DB2'):
            output_lines.extend(self._totals_lines())
        elif self.report_type.startswith('DB'):
            self._ensure_required_metrics()
            try:
                self.pubs.sort(
                    key=lambda x: METRICS[self.report_type].index(x.metric))
            except ValueError:
                pass

        for pub in sorted(self.pubs, key=lambda x: x.title):
            output_lines.append(pub.as_generic())

        return output_lines

    def _totals_lines(self):
        """Generate Totals for COUNTER report, as list of lists of cells."""
        total_lines = []
        metrics = set(resource.metric for resource in self.pubs)

        for metric in sorted(metrics):
            total_lines.append(self._totals_line(metric))

        return total_lines

    def _totals_line(self, metric):
        total_cells = [
            TOTAL_TEXT[self.report_type],
        ]
        publishers = set(resource.publisher for resource in self.pubs)
        if len(publishers) == 1:
            total_cells.append(publishers.pop())
        else:
            total_cells.append(u'')
        platforms = set(resource.platform for resource in self.pubs)
        if len(platforms) == 1:
            total_cells.append(platforms.pop())
        else:
            total_cells.append(u'')
        if self.report_type in ('JR1', 'BR1', 'BR2'):
            total_cells.extend([u''] * 4)
        elif self.report_type == 'DB2':
            total_cells.append(metric)
        total_usage = 0
        pdf_usage = 0
        html_usage = 0

        number_of_months = len(
            arrow.Arrow.range('month',
                              arrow.Arrow.fromdate(self.period[0]),
                              arrow.Arrow.fromdate(self.period[1])))
        month_data = [0] * number_of_months
        for pub in self.pubs:
            if pub.metric != metric:
                continue
            if self.report_type == 'JR1':
                pdf_usage += pub.pdf_total
                html_usage += pub.html_total
            for month, data in enumerate(pub):
                total_usage += data[2]
                month_data[month] += data[2]
        total_cells.append(six.text_type(total_usage))
        if self.report_type == 'JR1':
            total_cells.append(six.text_type(html_usage))
            total_cells.append(six.text_type(pdf_usage))
        total_cells.extend(six.text_type(d) for d in month_data)
        return total_cells

    def _table_header(self):
        """Generate header for COUNTER table for report, as list of cells."""
        header_cells = list(HEADER_FIELDS[self.report_type])
        for d_obj in arrow.Arrow.range('month',
                                       arrow.Arrow.fromdate(self.period[0]),
                                       arrow.Arrow.fromdate(self.period[1])):
            header_cells.append(d_obj.strftime('%b-%Y'))
        return header_cells

    def _ensure_required_metrics(self):
        """
        Build up a dict of sets of known metrics for each database.

        If any metric is missing add a 0 use
        :class:`CounterDatabase<CounterDatabase>`.
        Assumes platform and publisher are consistent across records.
        """
        try:
            required_metrics = METRICS[self.report_type]
        except LookupError:
            raise UnknownReportTypeError(self.report_type)

        dbs = collections.defaultdict(set)
        for database in self.pubs:
            dbs[database.title].add(database.metric)

        for database, metrics in six.iteritems(dbs):
            for metric in (m for m in required_metrics if m not in metrics):
                self.pubs.append(
                    CounterDatabase(
                        title=database,
                        platform=self.pubs[0].platform,
                        publisher=self.pubs[0].publisher,
                        period=self.period,
                        metric=metric,
                        month_data=[(self.period[0], 0), ]
                    ))


class CounterEresource(six.Iterator):
    """
    Base class for COUNTER statistics lines.

    Iterating returns (first_day_of_month, metric, usage) tuples.

    :param period: two-tuple of datetime.date objects corresponding
        to the beginning and end dates of the covered range

    :param metric: metric tracked by this report. Should be a value
        from pycounter.report.METRICS dict.

    :param month_data: a list containing usage data for this
        resource, as (datetime.date, usage) tuples

    :param title: title of the resource

    :param publisher: name of the resource's publisher

    :param platform: name of the platform providing the resource

    """

    def __init__(self, period=None, metric=None, month_data=None,
                 title="", platform="", publisher=""):
        self.period = period

        self.metric = metric
        self._full_data = []
        if month_data is not None:
            for item in month_data:
                self._full_data.append(item)

        self.title = title
        self.platform = platform
        self.publisher = publisher

    def __iter__(self):
        if self._full_data:
            for item in self._full_data:
                yield (item[0], self.metric, item[1])

    def _fill_months(self):
        """Ensure each month in period represented and zero fill if not."""
        start, end = self.period[0], self.period[1]
        try:
            for d_obj in arrow.Arrow.range('month',
                                           arrow.Arrow.fromdate(start),
                                           arrow.Arrow.fromdate(end)):
                if d_obj.date() not in (x[0] for x in self._full_data):
                    self._full_data.append((d_obj.date(), 0))
        except IndexError:
            pass
        else:
            self._full_data.sort()


class CounterJournal(CounterEresource):
    """
    Statistics for a single electronic journal.

    :param period: two-tuple of datetime.date objects corresponding
        to the beginning and end dates of the covered range

    :param metric: the metric tracked by this statistics line.
        (Should probably always be "FT Article Requests" for
        CounterJournal objects, as long as only JR1 is supported.)

    :param issn: eJournal's print ISSN

    :param eissn: eJournal's eISSN

    :param month_data: a list containing usage data for this
        journal, as (datetime.date, usage) tuples

    :param title: title of the resource

    :param publisher: name of the resource's publisher

    :param platform: name of the platform providing the resource

    :param html_total: total HTML usage for this title for reporting period

    :param pdf_total: total PDF usage for this title for reporting period

    """

    def __init__(self, period=None, metric=METRICS[u"JR1"],
                 issn=None, eissn=None, month_data=None,
                 title="", platform="", publisher="", html_total=0,
                 pdf_total=0, doi="", proprietary_id=""):
        super(CounterJournal, self).__init__(period, metric, month_data,
                                             title, platform, publisher)
        self.html_total = html_total
        self.pdf_total = pdf_total
        self.doi = doi
        self.proprietary_id = proprietary_id

        self.isbn = None

        if issn is not None:
            self.issn = issn
        else:
            self.issn = ''

        if eissn is not None:
            self.eissn = eissn
        else:
            self.eissn = ''

    def __repr__(self):
        return """<CounterJournal %s, publisher %s,
        platform %s>""" % (self.title, self.publisher, self.platform)

    def as_generic(self):
        """Get data for this line as list of COUNTER report cells."""
        self._fill_months()  # Ensure fill all months with zero at least
        data_line = [
            self.title,
            self.publisher,
            self.platform,
            self.doi,
            self.proprietary_id,
            self.issn,
            self.eissn,
        ]
        total_usage = 0
        month_data = []
        for data in self:
            total_usage += data[2]
            month_data.append(six.text_type(data[2]))
        data_line.append(six.text_type(total_usage))
        data_line.append(six.text_type(self.html_total))
        data_line.append(six.text_type(self.pdf_total))
        data_line.extend(month_data)
        return data_line


class CounterBook(CounterEresource):
    """
    statistics for a single electronic book.

    :ivar isbn: eBook's ISBN

    :ivar issn: eBook's ISSN (if any)

    :param month_data: a list containing usage data for this
        book, as (datetime.date, usage) tuples

    :param title: title of the resource

    :param publisher: name of the resource's publisher

    :param platform: name of the platform providing the resource

    """

    def __init__(self, period=None, metric=None, month_data=None,
                 title="", platform="", publisher="", isbn=None, issn=None,
                 doi="", proprietary_id="", print_isbn=None, online_isbn=None):
        super(CounterBook, self).__init__(period, metric, month_data,
                                          title, platform, publisher)
        self.eissn = None
        self.doi = doi
        self.proprietary_id = proprietary_id

        self._isbn = isbn
        self.print_isbn = print_isbn
        self.online_isbn = online_isbn

        if issn is not None:
            self.issn = issn
        else:
            self.issn = u''

    def __repr__(self):
        return """<CounterBook %s (ISBN: %s), publisher %s,
        platform %s>""" % (self.title, self.isbn, self.publisher,
                           self.platform)

    @property
    def isbn(self):
        """Return a suitable ISSN for the ebook.

        The tabular COUNTER reports only report an "ISBN", while the SUSHI
        (XML) reports include both a Print_ISBN and Online_ISBN.

         This property will return a generic ISBN given in the constructor,
         if any. If the CounterBook was created with no "isbn" but with
         online_ISBN and/or print_ISBN, the online one, if any, will be
         returned, otherwise the print.
        """
        return self._isbn or self.online_isbn or self.print_isbn or u''

    def as_generic(self):
        """Get data for this line as list of COUNTER report cells."""
        self._fill_months()  # Ensure fill all months with zero at least
        data_line = [
            self.title,
            self.publisher,
            self.platform,
            self.doi,
            self.proprietary_id,
            self.isbn,
            self.issn,
        ]
        total_usage = 0
        month_data = []
        for data in self:
            total_usage += data[2]
            month_data.append(six.text_type(data[2]))
        data_line.append(six.text_type(total_usage))
        data_line.extend(month_data)
        return data_line


class CounterDatabase(CounterEresource):
    """a COUNTER database report line."""

    def __init__(self, period=None, metric=None, month_data=None,
                 title="", platform="", publisher=""):
        super(CounterDatabase, self).__init__(period, metric, month_data,
                                              title, platform, publisher)
        self.isbn = None

    def as_generic(self):
        """Return data for this line as list of COUNTER report cells."""
        self._fill_months()

        data_line = [
            self.title,
            self.publisher,
            self.platform,
            self.metric,
        ]
        total_usage = 0
        month_data = []

        for data in self:
            total_usage += data[2]
            month_data.append(six.text_type(data[2]))

        data_line.append(six.text_type(total_usage))
        data_line.extend(month_data)

        return data_line


def parse(filename, filetype=None, encoding='utf-8',
          fallback_encoding='latin-1'):
    """Parse a COUNTER file, first attempting to determine type.

    Returns a :class:`CounterReport <CounterReport>` object.

    :param filename: path to COUNTER report to load and parse.
    :param filetype: type of file provided, one of "csv", "tsv", "xlsx".
        If set to None (the default), an attempt will be made to
        detect the correct type, first from the file extension, then from
        the file's contents.
    :param encoding: encoding to use to decode the file. Defaults to 'utf-8',
        ignored for XLSX files (which specify their encoding in their XML)
    :param fallback_encoding: alternative encoding to use to try to decode
        the file if the primary encoding fails. This defaults to 'latin-1',
        which will accept any bytes (possibly producing junk results...)
        Ignored for XLSX files.

    """
    if filetype is None:
        if filename.endswith('.tsv'):
            filetype = 'tsv'
        elif filename.endswith('.xlsx'):
            filetype = 'xlsx'
        elif filename.endswith('.csv'):
            filetype = 'csv'
        else:
            with open(filename, 'rb') as file_obj:
                filetype = guess_type_from_content(file_obj)

    if filetype == 'tsv':
        return parse_separated(filename, '\t', encoding, fallback_encoding)
    elif filetype == 'xlsx':
        return parse_xlsx(filename)
    elif filetype == 'csv':
        return parse_separated(filename, ',', encoding, fallback_encoding)
    else:
        raise PycounterException("Unknown file type %s" % filetype)


def parse_xlsx(filename):
    """Parse a COUNTER file in Excel format.

    Invoked automatically by ``parse``.

    :param filename: path to XLSX-format COUNTER report file.

    """
    from openpyxl import load_workbook
    with open(filename, 'rb') as xlsx_file:
        workbook = load_workbook(xlsx_file)
        worksheet = workbook[workbook.sheetnames[0]]
        row_it = worksheet.iter_rows()
        split_row_list = ([cell.value if cell.value is not None else ""
                           for cell in row] for row in row_it)

    return parse_generic(split_row_list)


def parse_separated(filename, delimiter, encoding='utf-8',
                    fallback_encoding='latin-1'):
    r"""Open COUNTER CSV/TSV report and parse into a CounterReport.

    Invoked automatically by :py:func:`parse`.

    :param filename: path to delimited COUNTER report file.

    :param delimiter: character (such as ',' or '\\t') used as the
        delimiter for this file

    :param encoding: file's encoding. Default: utf-8

    :param fallback_encoding: alternative encoding to try to decode if
        default fails. Throws a warning if used.

    :return: CounterReport object
    """
    with csvhelper.UnicodeReader(filename, delimiter=delimiter,
                                 fallback_encoding=fallback_encoding,
                                 encoding=encoding) as report_reader:
        return parse_generic(report_reader)


def parse_generic(report_reader):
    """Parse COUNTER report rows into a CounterReport.

    :param report_reader: a iterable object that yields lists COUNTER
        data formatted as tabular lists
    :return: CounterReport object

    """
    report = CounterReport()

    report.report_type, report.report_version = _get_type_and_version(
        six.next(report_reader)[0])

# noinspection PyTypeChecker
    report.metric = METRICS.get(report.report_type)

    report.customer = six.next(report_reader)[0]

    if report.report_version == 4:
        inst_id_line = six.next(report_reader)
        if inst_id_line:
            report.institutional_identifier = inst_id_line[0]
            if report.report_type == 'BR2':
                report.section_type = inst_id_line[1]

        six.next(report_reader)

        covered_line = six.next(report_reader)
        report.period = convert_covered(covered_line[0])

    six.next(report_reader)

    date_run_line = six.next(report_reader)
    report.date_run = convert_date_run(date_run_line[0])

    header = six.next(report_reader)

    try:
        report.year = _year_from_header(header, report)
    except AttributeError:
        warnings.warn("Could not determine year from malformed header")

    if report.report_version == 4:
        countable_header = header[0:8]
        for col in header[8:]:
            if col:
                countable_header.append(col)
        last_col = len(countable_header)
    else:
        last_col = 0
        for val in header:
            if 'YTD' in val:
                break
            last_col += 1

        start_date = datetime.date(report.year, 1, 1)
        end_date = last_day(convert_date_column(header[last_col - 1]))
        report.period = (start_date, end_date)

    if report.report_type != 'DB1':
        six.next(report_reader)

    if report.report_type == 'DB2':
        six.next(report_reader)

    for line in report_reader:
        if not line:
            continue
        report.pubs.append(_parse_line(line, report, last_col))

    return report


def _parse_line(line, report, last_col):
    """Parse a single line from a report.

    :param line: sequence of cells in a report line
    :param report: a CounterReport the line came from
    :param last_col: last column number containing data
    :return: an appropriate CounterResource subclass instance
    """
    issn = None
    eissn = None
    isbn = None
    html_total = 0
    pdf_total = 0
    doi = ""
    prop_id = ""

    if report.report_version == 4:
        if report.report_type.startswith('JR1'):
            old_line = line
            line = line[0:3] + line[5:7] + line[10:last_col]
            doi = old_line[3]
            prop_id = old_line[4]
            html_total = format_stat(old_line[8])
            pdf_total = format_stat(old_line[9])
            issn = line[3].strip()
            eissn = line[4].strip()

        elif report.report_type in ('BR1', 'BR2'):
            line = line[0:3] + line[5:7] + line[8:last_col]
            isbn = line[3].strip()
            issn = line[4].strip()

        elif report.report_type in ('DB1', 'DB2'):
            # format coincidentally works for these. This is a kludge
            # so leaving this explicit...
            pass
    else:
        if report.report_type.startswith('JR1'):
            html_total = format_stat(line[-2])
            pdf_total = format_stat(line[-1])
            issn = line[3].strip()
            eissn = line[4].strip()
        line = line[0:last_col]

    logging.debug(line)
    common_args = {
        'title': line[0],
        'publisher': line[1],
        'platform': line[2],
        'period': report.period
    }
    month_data = []
    curr_month = report.period[0]
    for data in line[5:]:
        month_data.append((curr_month, format_stat(data)))
        curr_month = next_month(curr_month)
    if report.report_type.startswith('JR'):
        return CounterJournal(metric=report.metric,
                              month_data=month_data,
                              doi=doi,
                              issn=issn,
                              eissn=eissn,
                              proprietary_id=prop_id,
                              html_total=html_total,
                              pdf_total=pdf_total,
                              **common_args
                              )
    elif report.report_type.startswith('BR'):
        return CounterBook(metric=report.metric,
                           month_data=month_data,
                           doi=doi,
                           issn=issn,
                           isbn=isbn,
                           proprietary_id=prop_id,
                           **common_args)
    elif report.report_type.startswith('DB'):
        return CounterDatabase(metric=line[3],
                               month_data=month_data,
                               **common_args)
    raise PycounterException("Should be unreachable")  # pragma: no cover


def _get_type_and_version(specifier):
    """Given a COUNTER report specifier, find the type and version.

    :param specifier: COUNTER report specifier
    :return: type, version tuple
    """
    report_types_clause = '|'.join(CODES)
    rt_match = re.match(
        r'.*(%s) Report (\d(?: GOA)?) ?\(R(\d)\)' % report_types_clause,
        specifier)
    if rt_match:
        report_type = (CODES[rt_match.group(1)] + rt_match.group(2))
        report_version = int(rt_match.group(3))
    else:
        raise UnknownReportTypeError("No match in line: %s" % specifier)
    if not any(report_type.startswith(x) for x in ('JR', 'BR', 'DB')):
        raise UnknownReportTypeError(report_type)

    return report_type, report_version


def _year_from_header(header, report):
    """Get the year for the report from the header.

    NOTE: for multi-year reports, this will be the date of the first month,
    and probably doesn't make sense to talk of a report having a year...
    """
    first_date_col = 10 if report.report_version == 4 else 5
    if report.report_type in ('BR1', 'BR2') and report.report_version == 4:
        first_date_col = 8
    elif report.report_type == 'DB1' and report.report_version == 4:
        first_date_col = 6
    elif report.report_type == 'DB2' and report.report_version == 4:
        first_date_col = 5
    year = int(header[first_date_col].split('-')[1])
    if year < 100:
        year += 2000

    return year



"""Helper functions used by pycounter."""
import calendar
import datetime

import six


def convert_covered(datestring):
    """
    Convert coverage period string to datetimes.

    :param datestring: the string to convert to a date. Format as
        'YYYY-MM-DD to YYYY-MM-DD'

    :return: tuple of datetime.date instances

    (Will also accept MM/DD/YYYY format, ISO 8601 timestamps, or existing
    datetime objects; these shouldn't be in COUNTER reports, but they
    do show up in real world data...)
    """
    start_string, end_string = datestring.split(" to ")
    start_date = convert_date_run(start_string)
    end_date = convert_date_run(end_string)

    return start_date, end_date


def convert_date_run(datestring):
    """
    Convert a date of the format 'YYYY-MM-DD' to a datetime.date object.

    (Will also accept MM/DD/YYYY format, ISO 8601 timestamps, or existing
    datetime objects; these shouldn't be in COUNTER reports, but they
    do show up in real world data...)

    :param datestring: the string to convert to a date.

    :return: datetime.date object

    """
    if isinstance(datestring, datetime.date):
        return datestring

    try:
        return datetime.datetime.strptime(datestring, "%Y-%m-%d").date()
    except ValueError:
        try:
            return datetime.datetime.strptime(datestring, "%m/%d/%Y").date()
        except ValueError:
            # ISO 8601 without timezone
            return datetime.datetime.strptime(datestring,
                                              "%Y-%m-%dT%H:%M:%S").date()


def convert_date_column(datestring):
    """
    Convert human-readable month to date of first day of month.

    :param datestring: the string to convert to a date. Format like "Jan-2014".

    :return: datetime.date
    """
    return datetime.datetime.strptime(datestring.strip(), "%b-%Y").date()


def last_day(orig_date):
    """
    Find last day of a month from any day in the month.

    :param orig_date: the date within the month for which we want the
        last day as datetime.date

    :return: datetime.date of last day of the month

    """
    day_number = calendar.monthrange(orig_date.year, orig_date.month)[1]
    return datetime.date(orig_date.year, orig_date.month, day_number)


def next_month(dateobj):
    """Find the first day of the next month after the given date.

    :param dateobj: the date within the month for which we want the
        next month's first day as datetime.date

    :return: datetime.date of the first day of the next month

    """
    year_delta, old_month = divmod(dateobj.month, 12)
    return datetime.date(dateobj.year + year_delta, old_month + 1, 1)


def prev_month(dateobj):
    """Find the first day of the previous month before the given date.

    :param dateobj: the date within the month for which we want the
        previous month's first day as datetime.date.

    :return: datetime.date of first day of the previous month.

    """
    year_delta, old_month = divmod(dateobj.month - 2, 12)
    return datetime.date(dateobj.year + year_delta, old_month + 1, 1)


def format_stat(stat):
    """Turn numbers possibly with embedded commas into integers.

    Also accepts existing ints, which may be pre-converted from Excel.

    :param stat: numeric value, possibly with commas

    :return: int
    """
    if isinstance(stat, six.integer_types):
        return stat

    stat = stat.replace(',', '')
    stat = stat.lstrip('=')
    try:
        return int(stat)
    except ValueError:
        return None


def guess_type_from_content(file_obj):
    """Guess type of a spreadsheet-like file.

    Defaults to assuming it's CSV, if it doesn't appear to be XLSX or TSV.

    :param file_obj: file-like object of which to determine type.

    :return: string, one of "xlsx", "tsv", "csv"
    """
    first_bytes = file_obj.read(2)
    if first_bytes == b"PK":
        filetype = 'xlsx'
    else:
        content = file_obj.read()
        if b'\t' in content:
            filetype = 'tsv'
        else:
            filetype = 'csv'
    return filetype



from __future__ import with_statement

import os
import platform

from setuptools import find_packages, setup

version = {}  # will be set by exec below

with open(os.path.join(
        os.path.dirname(__file__), 'pycounter/version.py'), 'r') as fp:
    exec(fp.read(), version)

with open('README.rst') as readmefile:
    readme = readmefile.read()

requirements = [
    'openpyxl',
    'requests',
    'six',
    'arrow',
    'click',
]

if platform.python_implementation() == "PyPy":
    requirements.append('lxml<=3.4.4')
else:
    requirements.append('lxml')

setup(
    name='pycounter',
    version=version['__version__'],
    packages=find_packages(),
    author='Health Sciences Library System, University of Pittsburgh',
    author_email='speargh@pitt.edu',
    maintainer='Geoffrey Spear',
    maintainer_email='speargh@pitt.edu',
    url='http://www.github.com/pitthsls/pycounter',
    description='Project COUNTER/NISO SUSHI statistics',
    long_description=readme,
    keywords='library COUNTER journals usage_statistics SUSHI',
    test_suite='pycounter.test',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        ],
    install_requires=requirements,
    tests_require=['httmock', 'mock', 'pytest'],
    entry_points={
        'console_scripts': ['sushiclient = pycounter.sushiclient:main']
    }
    )



"""Test output of JR1 report"""

from __future__ import absolute_import

import datetime
import logging
import os
import tempfile
import unittest

from pycounter import csvhelper
from pycounter import report


class TestOutputJR1(unittest.TestCase):
    """Test output of JR1"""
    def setUp(self):
        filename = os.path.join(os.path.dirname(__file__), 'data/C4JR1.csv')
        rep = report.parse(filename)
        with csvhelper.UnicodeReader(filename,
                                     delimiter=',') as report_reader:
            self.file_content = list(report_reader)

        self.output_content = rep.as_generic()

    def test_header_content(self):
        self.assertEqual(self.file_content[0:7], self.output_content[0:7])

    def test_table_header(self):
        self.assertEqual(self.file_content[7], self.output_content[7])

    def test_totals(self):
        self.assertEqual(self.file_content, self.output_content)

    def test_data(self):
        for index, line in enumerate(self.file_content[9:], 9):
            self.assertEqual(line, self.output_content[index])


class TestWritingJR1(unittest.TestCase):
    """Test write of JR1 to filesystem"""
    def setUp(self):
        self.filename = os.path.join(os.path.dirname(__file__),
                                     'data/simpleJR1.tsv')
        self.rep = report.parse(self.filename)

    def test_output_tsv(self):
        tf = tempfile.NamedTemporaryFile(delete=False)
        tf.close()
        tmp_loc = tf.name
        self.rep.write_tsv(tmp_loc)
        logging.debug("Temp file at %s", tmp_loc)

        with open(self.filename, 'rb') as orig_file:
            orig_content = orig_file.read()

        with open(tmp_loc, 'rb') as new_file:
            new_content = new_file.read()

        self.assertEqual(orig_content, new_content)


class TestOutputJR1Multi(unittest.TestCase):
    """Test output of JR1 - Multiple publishers and platforms."""
    def setUp(self):
        filename = os.path.join(os.path.dirname(__file__), 'data/C4JR1mul.csv')
        rep = report.parse(filename)
        with csvhelper.UnicodeReader(filename,
                                     delimiter=',') as report_reader:
            self.file_content = list(report_reader)

        self.output_content = rep.as_generic()

    def test_header_content(self):
        self.assertEqual(self.file_content[0:7], self.output_content[0:7])

    def test_table_header(self):
        self.assertEqual(self.file_content[7], self.output_content[7])

    def test_totals(self):
        self.assertEqual(self.file_content, self.output_content)

    def test_data(self):
        for index, line in enumerate(self.file_content[9:], 9):
            self.assertEqual(line, self.output_content[index])


class TestOrderingOutput(unittest.TestCase):
    """Test that titles are in alphabetical order when output"""
    def setUp(self):
        self.report = report.CounterReport(
            report_type='JR1',
            report_version=4,
            customer=u'Foo',
            institutional_identifier=u'Bar',
            period=(datetime.date(2016, 1, 1), datetime.date(2016, 1, 31))
        )
        self.report.pubs.append(
            report.CounterJournal(
                period=(datetime.date(2016, 1, 1), datetime.date(2016, 1, 31)),
                month_data=[(datetime.date(2016, 1, 1), 260)],
                title=u'Fake Journal',
                publisher=u'No one',
                platform=u'Your imagination'
            )
        )
        self.report.pubs.append(
            report.CounterJournal(
                period=(datetime.date(2016, 1, 1), datetime.date(2016, 1, 31)),
                month_data=[(datetime.date(2016, 1, 1), 62)],
                title=u'Another Fake Journal',
                publisher=u'No one',
                platform=u'Your imagination'
            )
        )
        self.output_content = self.report.as_generic()

    def test_ordering(self):
        self.assertEqual(self.output_content[9][0], u'Another Fake Journal')
        self.assertEqual(self.output_content[10][0], u'Fake Journal')



"""Constants used by pycounter."""

NS = {
    'SOAP-ENV': "http://schemas.xmlsoap.org/soap/envelope/",
    'sushi': "http://www.niso.org/schemas/sushi",
    'sushicounter': "http://www.niso.org/schemas/sushi/counter",
    'counter': "http://www.niso.org/schemas/counter",
    }

METRICS = {
    u"JR1": u"FT Article Requests",
    u"JR1 GOA": u"Gold Open Access Article Requests",
    u"BR1": u"Book Title Requests",
    u"BR2": u"Book Section Requests",
    u"DB1": [u"Regular Searches",
             u"Searches-federated and automated",
             u"Result Clicks",
             u"Record Views"],
    u"DB2": [u"Access denied: concurrent/simultaneous user license exceeded",
             u"Access denied: content item not licensed"]
}

DB_METRIC_MAP = {
    "search_reg": METRICS["DB1"][0],
    "search_fed": METRICS["DB1"][1],
    "result_click": METRICS["DB1"][2],
    "record_view": METRICS["DB1"][3],
    "turnaway": METRICS["DB2"][0],
    "no_license": METRICS["DB2"][1]
}

CODES = {
    u"Database": u"DB",
    u"Journal": u"JR",
    u"Book": u"BR",
    u"Title": u"TR",
    u"Platform": u"PR",
    u"Multimedia": u"MR",
    u"Consortium": u"CR",
}

# from http://www.niso.org/workrooms/sushi/registry/
# Not all of these are actually supported by pycounter
REPORT_DESCRIPTIONS = {
    u'BR1': u'Number of Successful Title Requests by Month and Title',
    u'BR2': u'Number of Successful Section Requests by Month and Title',
    u'BR3': u'Access Denied to Content Items by Month, Title, and Category',
    u'BR4': u'Access Denied to Content Items by Month, Platform, and Category',
    u'BR5': u'Total Searches by Month and Title',
    u'CR1': u'Number of Successful Full-text Journal Article or Book Chapter '
            u'Requests by Month',
    u'CR2': u'Total Searches by Month and Database',
    u'CR3': u'Number of Successful Multimedia Full Content Unit Requests '
            u'by Month and Collection',
    u'DB1': u'Total Searches, Result Clicks and Record Views by Month and '
            u'Database',
    u'DB2': u'Access Denied by Month, Database and Category',
    u'JR1': u'Number of Successful Full-Text Article Requests by Month and '
            u'Journal',
    u'JR1GOA': u'Number of Successful Gold Open Access Full-Text Article '
               u'Requests by Month and Journal',
    u'JR1a': u'Number of Successful Full-Text Article Requests from an '
             u'Archive by Month and Journal',
    u'JR2': u'Access Denied to Full Text Articles by Month, Journal, and '
            u'Category',
    u'JR3': u'Number of Successful Item Requests and Turnaways by Month, '
            u'Journal, and Page-Type',
    u'JR3mobile': u'Number of Successful Item Requests by Month, Journal, '
                  u'and Page-Type for usage on a mobile device',
    u'JR4': u'Total Searches Run by Month and Collection',
    u'JR5': u'Number of Successful Full-Text Article Requests by '
            u'Year-of-Publication (YOP) and Journal',
    u'MR1': u'Number of Successful Multimedia Full Content Unit Requests '
            u'by Month and Collection',
    u'MR2': u'Number of Successful Multimedia Full Content Unit Requests by '
            u'Month, Collection, and Item Type',
    u'PR1': u'Total Searches, Result Clicks, and Record Views by Month and '
            u'Platform',
    u'TR1': u'Number of Successful Requests for Journal Full-Text Articles '
            u'and Book Sections by Month and Title',
    u'TR1mobile': u'Number of Successful Requests for Journal Full-Text '
                  u'Articles and Book Sections by Month and Title '
                  u'(formatted for normal browsers/delivered to mobile '
                  u'devices AND formatted for mobile devices/delivered '
                  u'to mobile devices)',
    u'TR2': u'Access Denied to Full-Text Items by Month, Title, and Category',
    u'TR3': u'Number of Successful Item Requests by Month, Title, and '
            u'Page-Type',
    u'TR3mobile': u'Number of Successful Item Requests by Month, Title, '
                  u'and Page-Type (formatted for normal browsers/delivered '
                  u'to mobile devices and for mobile devices/delivered to '
                  u'mobile devices)',
}

HEADER_FIELDS = {
    "JR1": (
        u'Journal',
        u'Publisher',
        u'Platform',
        u'Journal DOI',
        u'Proprietary Identifier',
        u'Print ISSN',
        u'Online ISSN',
        u'Reporting Period Total',
        u'Reporting Period HTML',
        u'Reporting Period PDF',
    ),
    "JR2": (
        u'Journal',
        u'Publisher',
        u'Platform',
        u'Journal DOI',
        u'Proprietary Identifier',
        u'Print ISSN',
        u'Online ISSN',
        u'Reporting Period Total',
        u'Reporting Period HTML',
        u'Reporting Period PDF',
    ),
    "JR3": (
        u'Journal',
        u'Publisher',
        u'Platform',
        u'Journal DOI',
        u'Proprietary Identifier',
        u'Print ISSN',
        u'Online ISSN',
        u'Reporting Period Total',
        u'Reporting Period HTML',
        u'Reporting Period PDF',
    ),
    "BR1": (
        u'',
        u'Publisher',
        u'Platform',
        u'Book DOI',
        u'Proprietary Identifier',
        u'ISBN',
        u'ISSN',
        u'Reporting Period Total',
    ),
    "BR2": (
        u'',
        u'Publisher',
        u'Platform',
        u'Book DOI',
        u'Proprietary Identifier',
        u'ISBN',
        u'ISSN',
        u'Reporting Period Total',
    ),
    "BR3": (
        u'',
        u'Publisher',
        u'Platform',
        u'Book DOI',
        u'Proprietary Identifier',
        u'ISBN',
        u'ISSN',
        u'Reporting Period Total',
    ),
    "DB1": (
        u'Database',
        u'Publisher',
        u'Platform',
        u'User Activity',
        u'Reporting Period Total',
    ),
    "DB2": (
        u'Database',
        u'Publisher',
        u'Platform',
        u'Access denied category',
        u'Reporting Period Total',
    ),
}

TOTAL_TEXT = {
    'JR1': u'Total for all journals',
    'BR1': u'Total for all titles',
    'BR2': u'Total for all titles',
    'DB2': u'Total for all databases',
}



"""regression test to ensure all output is of text type"""

from __future__ import absolute_import

import os

import pytest

from six import text_type

from pycounter import report


@pytest.mark.parametrize('attribute', ['title', 'publisher', 'platform'])
@pytest.mark.parametrize('filename',
                         ['JR1.xlsx', 'simpleJR1.tsv', 'simpleJR1.csv'])
def test_unicode_fields(filename, attribute):
    """All parsers should return text fields as unicode"""
    rep = report.parse(os.path.join(os.path.dirname(__file__),
                                    'data', filename))
    assert isinstance(getattr(rep.pubs[0], attribute), text_type)



"""pycounter: Project COUNTER/NISO SUSHI statistics."""
from __future__ import absolute_import

from pycounter import exceptions, report, sushi
from pycounter.version import __version__

__all__ = ('__version__', 'report', 'sushi', 'exceptions')
